@@ -7,7 +7,6 @@ from binascii import a2b_hex
 from Cryptodome.Cipher import AES
 from Cryptodome.Cipher._mode_gcm import GcmMode
 from struct import Struct
-from io import BytesIO, SEEK_END
 from socket import socket
 import socket as s
 import threading
@@ -58,7 +57,7 @@ S_ESTABLISHED = b'\x03'
 # typing
 _Address = Tuple[Any, ...]
 _WildAddress = Union[_Address, str, bytes]
-_BroadcastHook = Callable[['Packet'], None]
+_BroadcastHook = Callable[['Packet', 'SecureReliableSocket'], None]
 
 
 class CycInt(int):
@@ -186,6 +185,7 @@ class SecureReliableSocket(socket):
         self.receiver_unread_size = 0
         self.receiver_buffer = get_self_bind_sock()
         # broadcast hook
+        # note: don't block this method or backend thread will be broken
         self.broadcast_hook_fnc: Optional[_BroadcastHook] = None
         # status
         self.loss = 0
@@ -339,7 +339,7 @@ class SecureReliableSocket(socket):
         log.debug("success get MUT size %db", self.mtu_size)
 
         # success establish connection
-        threading.Thread(target=self._backend, daemon=True).start()
+        threading.Thread(target=self._backend, name="SRUDP", daemon=True).start()
         self.established = True
 
         # auto exit when program closed
@@ -385,7 +385,6 @@ class SecureReliableSocket(socket):
         temporary = dict()
         retransmit_packets: Deque[Packet] = deque()
         retransmitted: Deque[float] = deque(maxlen=16)
-        broadcast_packets: Deque[Packet] = deque(maxlen=16)
         last_packet: Optional[Packet] = None
         last_receive_time = time()
         last_ack_time = time()
@@ -478,11 +477,12 @@ class SecureReliableSocket(socket):
             # broadcast packet
             if packet.control & CONTROL_BCT:
                 if self.broadcast_hook_fnc is not None:
-                    self.broadcast_hook_fnc(packet)
+                    self.broadcast_hook_fnc(packet, self)
                 elif last_packet is None or last_packet.control & CONTROL_EOF:
                     self._push_receive_buffer(packet.data)
                 else:
-                    broadcast_packets.append(packet)
+                    # note: acquire realtime response
+                    log.debug("throw away %s", packet)
                 continue
 
             """normal packet from here (except PSH, EOF)"""
@@ -557,14 +557,14 @@ class SecureReliableSocket(socket):
 
             # reached EOF & push broadcast packets
             if packet.control & CONTROL_EOF:
-                for p in broadcast_packets:
-                    self._push_receive_buffer(p.data)
-                broadcast_packets.clear()
+                # note: stopped sending broadcast packet after main stream for realtime
+                log.debug("reached EOF of data")
 
             # update last packet
             last_packet = packet
 
         # close
+        log.debug("srudp socket is closing now")
         self.close()
 
     def _push_receive_buffer(self, data: bytes) -> None:
@@ -649,9 +649,13 @@ class SecureReliableSocket(socket):
         elif not self.established:
             return b""
         else:
-            data = self.receiver_buffer.recv(buflen, flags)
-            self.receiver_unread_size -= len(data)
-            return data
+            try:
+                data = self.receiver_buffer.recv(buflen, flags)
+                self.receiver_unread_size -= len(data)
+                return data
+            except ConnectionError:
+                # self.close() called and buffer closed
+                return b""
 
     def _encrypt(self, data: bytes) -> bytes:
         """encrypt by AES-GCM (more secure than CBC mode)"""
@@ -679,6 +683,23 @@ class SecureReliableSocket(socket):
         """change inner receiver's instead"""
         self.receiver_buffer.settimeout(value)
 
+    def getblocking(self) -> bool:
+        """return inner receiver's instead"""
+        return self.receiver_buffer.getblocking()
+
+    def setblocking(self, flag: bool) -> None:
+        """change inner receiver's instead"""
+        self.receiver_buffer.setblocking(flag)
+
+    def getpeername(self) -> _Address:
+        """connection info or raise OSError"""
+        if self.is_closed:
+            raise OSError("socket is closed")
+        elif self.address is None:
+            raise OSError("not found peer connection")
+        else:
+            return self.address
+
     @property
     def is_closed(self) -> bool:
         if self.fileno() == -1:
@@ -765,7 +786,7 @@ def main() -> None:
                 log.debug("send broadcast!")
             sleep(20)
 
-    def broadcast_hook(packet: Packet) -> None:
+    def broadcast_hook(packet: Packet, _sock: SecureReliableSocket) -> None:
         log.debug("find you!!! (%s)", packet)
 
     sock.broadcast_hook_fnc = broadcast_hook
